@@ -1,11 +1,9 @@
-// kimgio module for SGI images
-//
-// Copyright (C) 2004  Melchior FRANZ  <mfranz@kde.org>
-//
-// This program is free software; you can redistribute it and/or
-// modify it under the terms of the Lesser GNU General Public License as
-// published by the Free Software Foundation; either version 2 of the
-// License, or (at your option) any later version.
+/*
+    kimgio module for SGI images
+    SPDX-FileCopyrightText: 2004 Melchior FRANZ <mfranz@kde.org>
+
+    SPDX-License-Identifier: LGPL-2.0-or-later
+*/
 
 /* this code supports:
  * reading:
@@ -22,18 +20,22 @@
  */
 
 #include "rgb_p.h"
+#include "util_p.h"
 
+#include <QList>
 #include <QMap>
-#include <QVector>
 
+#include <QDebug>
 #include <QImage>
-// #include <QDebug>
 
-class RLEData : public QVector<uchar>
+class RLEData : public QList<uchar>
 {
 public:
-    RLEData() {}
-    RLEData(const uchar *d, uint l, uint o) : _offset(o)
+    RLEData()
+    {
+    }
+    RLEData(const uchar *d, uint l, uint o)
+        : _offset(o)
     {
         for (uint i = 0; i < l; i++) {
             append(d[i]);
@@ -53,9 +55,13 @@ private:
 class RLEMap : public QMap<RLEData, uint>
 {
 public:
-    RLEMap() : _counter(0), _offset(0) {}
+    RLEMap()
+        : _counter(0)
+        , _offset(0)
+    {
+    }
     uint insert(const uchar *d, uint l);
-    QVector<const RLEData *> vector();
+    QList<const RLEData *> vector();
     void setBaseOffset(uint o)
     {
         _offset = o;
@@ -76,7 +82,12 @@ public:
     bool writeImage(const QImage &);
 
 private:
-    enum { NORMAL, DITHERED, SCREEN, COLORMAP }; // colormap
+    enum {
+        NORMAL,
+        DITHERED,
+        SCREEN,
+        COLORMAP,
+    }; // colormap
     QIODevice *_dev;
     QDataStream _stream;
 
@@ -96,7 +107,7 @@ private:
     QByteArray _data;
     QByteArray::Iterator _pos;
     RLEMap _rlemap;
-    QVector<const RLEData *> _rlevector;
+    QList<const RLEData *> _rlevector;
     uint _numrows;
 
     bool readData(QImage &);
@@ -110,9 +121,9 @@ private:
     uchar intensity(uchar);
 };
 
-SGIImage::SGIImage(QIODevice *io) :
-    _starttab(nullptr),
-    _lengthtab(nullptr)
+SGIImage::SGIImage(QIODevice *io)
+    : _starttab(nullptr)
+    , _lengthtab(nullptr)
 {
     _dev = io;
     _stream.setDevice(_dev);
@@ -128,7 +139,8 @@ SGIImage::~SGIImage()
 
 bool SGIImage::getRow(uchar *dest)
 {
-    int n, i;
+    int n;
+    int i;
     if (!_rle) {
         for (i = 0; i < _xsize; i++) {
             if (_pos >= _data.end()) {
@@ -144,13 +156,16 @@ bool SGIImage::getRow(uchar *dest)
         if (_bpc == 2) {
             _pos++;
         }
+        if (_pos >= _data.end()) {
+            return false;
+        }
         n = *_pos & 0x7f;
         if (!n) {
             break;
         }
 
         if (*_pos++ & 0x80) {
-            for (; i < _xsize && n--; i++) {
+            for (; i < _xsize && _pos < _data.end() && n--; i++) {
                 *dest++ = *_pos;
                 _pos += _bpc;
             }
@@ -171,7 +186,8 @@ bool SGIImage::readData(QImage &img)
     quint32 *start = _starttab;
     QByteArray lguard(_xsize, 0);
     uchar *line = (uchar *)lguard.data();
-    unsigned x, y;
+    unsigned x;
+    unsigned y;
 
     if (!_rle) {
         _pos = _data.begin();
@@ -248,7 +264,7 @@ bool SGIImage::readImage(QImage &img)
     qint16 u16;
     qint32 u32;
 
-//     qDebug() << "reading rgb ";
+    //     qDebug() << "reading rgb ";
 
     // magic
     _stream >> u16;
@@ -258,42 +274,42 @@ bool SGIImage::readImage(QImage &img)
 
     // verbatim/rle
     _stream >> _rle;
-//     qDebug() << (_rle ? "RLE" : "verbatim");
+    //     qDebug() << (_rle ? "RLE" : "verbatim");
     if (_rle > 1) {
         return false;
     }
 
     // bytes per channel
     _stream >> _bpc;
-//     qDebug() << "bytes per channel: " << int(_bpc);
-    if (_bpc == 1)
+    //     qDebug() << "bytes per channel: " << int(_bpc);
+    if (_bpc == 1) {
         ;
-    else if (_bpc == 2) {
-//         qDebug() << "dropping least significant byte";
+    } else if (_bpc == 2) {
+        //         qDebug() << "dropping least significant byte";
     } else {
         return false;
     }
 
     // number of dimensions
     _stream >> _dim;
-//     qDebug() << "dimensions: " << _dim;
+    //     qDebug() << "dimensions: " << _dim;
     if (_dim < 1 || _dim > 3) {
         return false;
     }
 
     _stream >> _xsize >> _ysize >> _zsize >> _pixmin >> _pixmax >> u32;
-//     qDebug() << "x: " << _xsize;
-//     qDebug() << "y: " << _ysize;
-//     qDebug() << "z: " << _zsize;
+    //     qDebug() << "x: " << _xsize;
+    //     qDebug() << "y: " << _ysize;
+    //     qDebug() << "z: " << _zsize;
 
     // name
     _stream.readRawData(_imagename, 80);
     _imagename[79] = '\0';
 
     _stream >> _colormap;
-//     qDebug() << "colormap: " << _colormap;
+    //     qDebug() << "colormap: " << _colormap;
     if (_colormap != NORMAL) {
-        return false;    // only NORMAL supported
+        return false; // only NORMAL supported
     }
 
     for (int i = 0; i < 404; i++) {
@@ -301,7 +317,7 @@ bool SGIImage::readImage(QImage &img)
     }
 
     if (_dim == 1) {
-//         qDebug() << "1-dimensional images aren't supported yet";
+        //         qDebug() << "1-dimensional images aren't supported yet";
         return false;
     }
 
@@ -309,16 +325,29 @@ bool SGIImage::readImage(QImage &img)
         return false;
     }
 
-    _numrows = _ysize * _zsize;
+    img = imageAlloc(_xsize, _ysize, QImage::Format_RGB32);
+    if (img.isNull()) {
+        qWarning() << "Failed to allocate image, invalid dimensions?" << QSize(_xsize, _ysize);
+        return false;
+    }
 
-    img = QImage(_xsize, _ysize, QImage::Format_RGB32);
+    if (_zsize == 0) {
+        return false;
+    }
 
     if (_zsize == 2 || _zsize == 4) {
         img = img.convertToFormat(QImage::Format_ARGB32);
     } else if (_zsize > 4) {
-//         qDebug() << "using first 4 of " << _zsize << " channels";
+        //         qDebug() << "using first 4 of " << _zsize << " channels";
+        // Only let this continue if it won't cause a int overflow later
+        // this is most likely a broken file anyway
+        if (_ysize > std::numeric_limits<int>::max() / _zsize) {
+            return false;
+        }
     }
 
+    _numrows = _ysize * _zsize;
+
     if (_rle) {
         uint l;
         _starttab = new quint32[_numrows];
@@ -326,6 +355,9 @@ bool SGIImage::readImage(QImage &img)
             _stream >> _starttab[l];
             _starttab[l] -= 512 + _numrows * 2 * sizeof(quint32);
         }
+        for (; l < _numrows; l++) {
+            _starttab[l] = 0;
+        }
 
         _lengthtab = new quint32[_numrows];
         for (l = 0; l < _numrows; l++) {
@@ -336,16 +368,18 @@ bool SGIImage::readImage(QImage &img)
     _data = _dev->readAll();
 
     // sanity check
-    if (_rle)
-        for (uint o = 0; o < _numrows; o++)
+    if (_rle) {
+        for (uint o = 0; o < _numrows; o++) {
             // don't change to greater-or-equal!
             if (_starttab[o] + _lengthtab[o] > (uint)_data.size()) {
-//                 qDebug() << "image corrupt (sanity check failed)";
+                //                 qDebug() << "image corrupt (sanity check failed)";
                 return false;
             }
+        }
+    }
 
     if (!readData(img)) {
-//         qDebug() << "image corrupt (incomplete scanline)";
+        //         qDebug() << "image corrupt (incomplete scanline)";
         return false;
     }
 
@@ -363,7 +397,8 @@ void RLEData::write(QDataStream &s)
 
 bool RLEData::operator<(const RLEData &b) const
 {
-    uchar ac, bc;
+    uchar ac;
+    uchar bc;
     for (int i = 0; i < qMin(size(), b.size()); i++) {
         ac = at(i);
         bc = b[i];
@@ -386,9 +421,9 @@ uint RLEMap::insert(const uchar *d, uint l)
     return QMap<RLEData, uint>::insert(data, _counter++).value();
 }
 
-QVector<const RLEData *> RLEMap::vector()
+QList<const RLEData *> RLEMap::vector()
 {
-    QVector<const RLEData *> v(size());
+    QList<const RLEData *> v(size());
     for (Iterator it = begin(); it != end(); ++it) {
         v.replace(it.value(), &it.key());
     }
@@ -409,8 +444,13 @@ uchar SGIImage::intensity(uchar c)
 
 uint SGIImage::compact(uchar *d, uchar *s)
 {
-    uchar *dest = d, *src = s, patt, *t, *end = s + _xsize;
-    int i, n;
+    uchar *dest = d;
+    uchar *src = s;
+    uchar patt;
+    uchar *t;
+    uchar *end = s + _xsize;
+    int i;
+    int n;
     while (src < end) {
         for (n = 0, t = src; t + 2 < end && !(*t == t[1] && *t == t[2]); t++) {
             n++;
@@ -453,11 +493,19 @@ bool SGIImage::scanData(const QImage &img)
     uchar *line = (uchar *)lineguard.data();
     uchar *buf = (uchar *)bufguard.data();
     const QRgb *c;
-    unsigned x, y;
+    unsigned x;
+    unsigned y;
     uint len;
 
     for (y = 0; y < _ysize; y++) {
-        c = reinterpret_cast<const QRgb *>(img.scanLine(_ysize - y - 1));
+        const int yPos = _ysize - y - 1; // scanline doesn't do any sanity checking
+        if (yPos >= img.height()) {
+            qWarning() << "Failed to get scanline for" << yPos;
+            return false;
+        }
+
+        c = reinterpret_cast<const QRgb *>(img.scanLine(yPos));
+
         for (x = 0; x < _xsize; x++) {
             buf[x] = intensity(qRed(*c++));
         }
@@ -471,7 +519,13 @@ bool SGIImage::scanData(const QImage &img)
 
     if (_zsize != 2) {
         for (y = 0; y < _ysize; y++) {
-            c = reinterpret_cast<const QRgb *>(img.scanLine(_ysize - y - 1));
+            const int yPos = _ysize - y - 1;
+            if (yPos >= img.height()) {
+                qWarning() << "Failed to get scanline for" << yPos;
+                return false;
+            }
+
+            c = reinterpret_cast<const QRgb *>(img.scanLine(yPos));
             for (x = 0; x < _xsize; x++) {
                 buf[x] = intensity(qGreen(*c++));
             }
@@ -480,7 +534,13 @@ bool SGIImage::scanData(const QImage &img)
         }
 
         for (y = 0; y < _ysize; y++) {
-            c = reinterpret_cast<const QRgb *>(img.scanLine(_ysize - y - 1));
+            const int yPos = _ysize - y - 1;
+            if (yPos >= img.height()) {
+                qWarning() << "Failed to get scanline for" << yPos;
+                return false;
+            }
+
+            c = reinterpret_cast<const QRgb *>(img.scanLine(yPos));
             for (x = 0; x < _xsize; x++) {
                 buf[x] = intensity(qBlue(*c++));
             }
@@ -494,7 +554,13 @@ bool SGIImage::scanData(const QImage &img)
     }
 
     for (y = 0; y < _ysize; y++) {
-        c = reinterpret_cast<const QRgb *>(img.scanLine(_ysize - y - 1));
+        const int yPos = _ysize - y - 1;
+        if (yPos >= img.height()) {
+            qWarning() << "Failed to get scanline for" << yPos;
+            return false;
+        }
+
+        c = reinterpret_cast<const QRgb *>(img.scanLine(yPos));
         for (x = 0; x < _xsize; x++) {
             buf[x] = intensity(qAlpha(*c++));
         }
@@ -527,7 +593,7 @@ void SGIImage::writeHeader()
 void SGIImage::writeRle()
 {
     _rle = 1;
-//     qDebug() << "writing RLE data";
+    //     qDebug() << "writing RLE data";
     writeHeader();
     uint i;
 
@@ -550,11 +616,12 @@ void SGIImage::writeRle()
 void SGIImage::writeVerbatim(const QImage &img)
 {
     _rle = 0;
-//     qDebug() << "writing verbatim data";
+    //     qDebug() << "writing verbatim data";
     writeHeader();
 
     const QRgb *c;
-    unsigned x, y;
+    unsigned x;
+    unsigned y;
 
     for (y = 0; y < _ysize; y++) {
         c = reinterpret_cast<const QRgb *>(img.scanLine(_ysize - y - 1));
@@ -597,7 +664,7 @@ void SGIImage::writeVerbatim(const QImage &img)
 
 bool SGIImage::writeImage(const QImage &image)
 {
-//     qDebug() << "writing "; // TODO add filename
+    //     qDebug() << "writing "; // TODO add filename
     QImage img = image;
     if (img.allGray()) {
         _dim = 2, _zsize = 1;
@@ -605,19 +672,31 @@ bool SGIImage::writeImage(const QImage &image)
         _dim = 3, _zsize = 3;
     }
 
-    if (img.format() == QImage::Format_ARGB32) {
+    auto hasAlpha = img.hasAlphaChannel();
+    if (hasAlpha) {
         _dim = 3, _zsize++;
     }
 
-    img = img.convertToFormat(QImage::Format_RGB32);
+    if (hasAlpha && img.format() != QImage::Format_ARGB32) {
+        img = img.convertToFormat(QImage::Format_ARGB32);
+    } else if (!hasAlpha && img.format() != QImage::Format_RGB32) {
+        img = img.convertToFormat(QImage::Format_RGB32);
+    }
     if (img.isNull()) {
-//         qDebug() << "can't convert image to depth 32";
+        //         qDebug() << "can't convert image to depth 32";
+        return false;
+    }
+
+    const int w = img.width();
+    const int h = img.height();
+
+    if (w > 65535 || h > 65535) {
         return false;
     }
 
     _bpc = 1;
-    _xsize = img.width();
-    _ysize = img.height();
+    _xsize = w;
+    _ysize = h;
     _pixmin = ~0u;
     _pixmax = 0;
     _colormap = NORMAL;
@@ -626,7 +705,7 @@ bool SGIImage::writeImage(const QImage &image)
     _rlemap.setBaseOffset(512 + _numrows * 2 * sizeof(quint32));
 
     if (!scanData(img)) {
-//         qDebug() << "this can't happen";
+        //         qDebug() << "this can't happen";
         return false;
     }
 
@@ -638,12 +717,6 @@ bool SGIImage::writeImage(const QImage &image)
         rle_size += _rlevector[i]->size();
     }
 
-//     qDebug() << "minimum intensity: " << _pixmin;
-//     qDebug() << "maximum intensity: " << _pixmax;
-//     qDebug() << "saved scanlines: " << _numrows - _rlemap.size();
-//     qDebug() << "total savings: " << (verbatim_size - rle_size) << " bytes";
-//     qDebug() << "compression: " << (rle_size * 100.0 / verbatim_size) << '%';
-
     if (verbatim_size <= rle_size) {
         writeVerbatim(img);
     } else {
@@ -699,15 +772,14 @@ bool RGBHandler::canRead(QIODevice *device)
         device->seek(oldPos);
     }
 
-    return head.size() >= 4 && head.startsWith("\x01\xda\x01") && (head[3] == 1 || head[3] == 2);
+    return head.size() >= 4 && head.startsWith("\x01\xda") && (head[2] == 0 || head[2] == 1) && (head[3] == 1 || head[3] == 2);
 }
 
 ///////////////////////////////////////////////////////////////////////////////
 
 QImageIOPlugin::Capabilities RGBPlugin::capabilities(QIODevice *device, const QByteArray &format) const
 {
-    if (format == "rgb" || format ==  "rgba" ||
-            format ==  "bw" || format == "sgi") {
+    if (format == "rgb" || format == "rgba" || format == "bw" || format == "sgi") {
         return Capabilities(CanRead | CanWrite);
     }
     if (!format.isEmpty()) {
@@ -734,3 +806,5 @@ QImageIOHandler *RGBPlugin::create(QIODevice *device, const QByteArray &format)
     handler->setFormat(format);
     return handler;
 }
+
+#include "moc_rgb_p.cpp"
